@@ -8,7 +8,7 @@ use List::AllUtils 'each_array', 'indexes';
 use Storable 'dclone';
 use Carp 'croak';
 
-our $VERSION = "v0.0.16";
+our $VERSION = "v0.0.17";
 
 has bs_prefix => sub { die "bs_prefix is required" };
 has c => sub { die "c is required" };
@@ -47,7 +47,7 @@ sub lock_state ($self, $channel_names, $sub, %opts) {
     my $notify_payload_size_limit = $self->notify_payload_size_limit;
     my $event_patch_sequence_name = $self->event_patch_sequence_name;
 
-    my $multi_mode = length ref $channel_names;
+    my $multi_mode = ref $channel_names eq 'ARRAY';
     $channel_names = [$channel_names] if not $multi_mode;
     my $db = $self->db;
     my $tx = $self->pg_tx;
