@@ -25,11 +25,12 @@ use Encode 'encode_utf8';
 
 use experimental 'postderef';
 
-our $VERSION = "v0.0.16";
+our $VERSION = "v0.0.17";
 
 use constant {
+    DEFAULT_CLEANUP_INTERVAL          => 5,
     DEFAULT_HEARTBEAT_INTERVAL        => 5,
-    DEFAULT_HEARTBEAT_TIMEOUT         => 10,
+    DEFAULT_HEARTBEAT_TIMEOUT         => 5,
     DEFAULT_PREFIX                    => 'bs',
     DEFAULT_PING_INTERVAL             => 15,
     DEFAULT_NOTIFY_PAYLOAD_SIZE_LIMIT => 8000,
@@ -37,11 +38,13 @@ use constant {
 };
 
 my $WORKERS_CHANNEL = '_bs:workers';
+my $CLEANUP_CHANNEL = '_bs:cleanup';
 
 sub register ($self, $app, $config) {
     my $worker_uuid = Data::GUID->new->as_base64;
 
     # config params
+    my $cleanup_interval = $config->{cleanup_interval} // DEFAULT_CLEANUP_INTERVAL;
     my $heartbeat_interval = $config->{heartbeat_interval} // DEFAULT_HEARTBEAT_INTERVAL;
     my $heartbeat_timeout = $config->{heartbeat_timeout} // DEFAULT_HEARTBEAT_TIMEOUT;
     my $bs_prefix = $config->{prefix} // DEFAULT_PREFIX;
@@ -71,7 +74,7 @@ sub register ($self, $app, $config) {
     my $last_heartbeat_dt;
 
     my sub calculate_hb_expiry_dur () {
-        state $hb_e_d = $heartbeat_interval + max($heartbeat_interval * 0.5, 10);
+        state $hb_e_d = $heartbeat_interval + $heartbeat_timeout;
     }
 
     # on workers startup:
@@ -190,6 +193,29 @@ sub register ($self, $app, $config) {
         )->subscribe(sub {
             Mojo::IOLoop->stop_gracefully;
         });
+
+        # one process needs to call "global_cleanup" every "cleanup_interval" seconds:
+        if ($cleanup_interval) {
+            rx_timer(rand($cleanup_interval), $cleanup_interval + 0.1 + rand())->subscribe(sub {
+                my $t0 = Time::HiRes::time();
+                return unless
+                    $t0 >= $app->$bs_prefix->get_state($CLEANUP_CHANNEL)->{last_cleanup} + $cleanup_interval;
+
+                my $me_is_cleanup_executor = 0;
+                $app->$bs_prefix->lock_state($CLEANUP_CHANNEL, sub ($state) {
+                    if (Time::HiRes::time() >= $state->{last_cleanup} + $cleanup_interval) {
+                        $state->{last_cleanup} = $t0;
+                        $me_is_cleanup_executor = 1;
+                        return undef, $state;
+                    } else {
+                        return undef, undef;
+                    }
+                });
+                return unless $me_is_cleanup_executor;
+
+                $app->$bs_prefix->global_cleanup;
+            });
+        }
     });
 
     $app->helper("$bs_prefix.on_action", sub ($c, $channel_name, $action_name, $action_sub) {
@@ -521,73 +547,81 @@ sub register ($self, $app, $config) {
         my $workers_state = $c->$bs_prefix->get_state($WORKERS_CHANNEL);
         my $time = get_time;
         foreach my $worker (values %$workers_state) {
-            if ($time > $worker->{heartbeat} + calculate_hb_expiry_dur()) {
-                my $db = $c->$bs_prefix->db;
+            next if $time <= $worker->{heartbeat} + calculate_hb_expiry_dur();
 
-                # 0. set 'banned' property of worker to true
-                $c->$bs_prefix->lock_state($WORKERS_CHANNEL, sub ($state) {
-                    $state->{$worker->{uuid}}{banned} = true
-                        if exists $state->{$worker->{uuid}};
-                    return undef, $state;
-                }, no_ban => 1);
+            my $db = $c->$bs_prefix->db;
 
-                # 1. find channel names from "guards" table
-                my $worker_uuid = $worker->{uuid};
-                my @channels = $db->query(q{
-                    SELECT c.*
-                        FROM guards g JOIN channel c ON (c.id = g.channel_id)
-                        WHERE g.worker_uuid = ?
-                }, $worker_uuid)->hashes->@*;
+            # 0. set 'banned' property of worker to true
+            $c->$bs_prefix->lock_state($WORKERS_CHANNEL, sub ($state) {
+                $state->{$worker->{uuid}}{banned} = true
+                    if exists $state->{$worker->{uuid}};
+                return undef, $state;
+            }, no_ban => 1);
 
-                # 2. call cleanup subs (deleting guards rows in the process)
-                CHANNEL:
-                foreach my $channel (@channels) {
-                    my $txn = $c->$bs_prefix->begin;
-                    my $db = $txn->db;
-                    my $guards_counter = $db->select('guards',
-                        ['counter'],
-                        {
-                            worker_uuid => $worker_uuid,
-                            channel_id  => $channel->{id},
-                        },
-                        { for => 'update' },
-                    )->hash->{counter};
-                    defined $guards_counter or next CHANNEL;
+            # 1. find channel names from "guards" table
+            my $worker_uuid = $worker->{uuid};
+            my @channels = $db->query(q{
+                SELECT c.*
+                    FROM guards g JOIN channel c ON (c.id = g.channel_id)
+                    WHERE g.worker_uuid = ?
+            }, $worker_uuid)->hashes->@*;
 
-                    # call cleanup sub
-                    my $channel_name = $channel->{name};
-                    if (my $cleanup_sub = BoardStreams::Registry->get_cleanup($channel_name)) {
-                        $c->$bs_prefix->lock_state($channel_name, sub ($state) {
-                            my $alive_workers = $c->$bs_prefix->alive_workers;
-                            my $new_state = $cleanup_sub->($channel_name, $state, $alive_workers);
+            # 2. call cleanup subs (deleting guards rows in the process)
+            CHANNEL:
+            foreach my $channel (@channels) {
+                my $txn = $c->$bs_prefix->begin;
+                my $db = $txn->db;
+                my $channel_name = $channel->{name};
 
-                            return undef, $new_state;
-                        }, no_ban => 1);
-                    }
+                # lock channel to prevent lock_state from modifying guard counts
+                $db->select('channel',
+                    'id',
+                    { name => $channel_name },
+                    { for => 'update' },
+                );
+                my $guard_row = $db->select('guards',
+                    ['counter'],
+                    {
+                        worker_uuid => $worker_uuid,
+                        channel_id  => $channel->{id},
+                    },
+                    { for => 'update' },
+                )->hash;
+                defined $guard_row or next CHANNEL;
+                my $guards_counter = $guard_row->{counter};
 
-                    $db->update('guards',
-                        { counter => \"counter - $guards_counter" },
-                        {
-                            worker_uuid => $worker_uuid,
-                            channel_id => $channel->{id},
-                        },
-                    );
-                    $db->delete('guards',
-                        {
-                            worker_uuid => $worker_uuid,
-                            channel_id  => $channel->{id},
-                            counter     => {'<=', 0},
-                        },
-                    );
-                    $txn->commit;
+                # call cleanup sub
+                if (my $cleanup_sub = BoardStreams::Registry->get_cleanup($channel_name)) {
+                    $txn->lock_state($channel_name, sub ($state) {
+                        my $alive_workers = $c->$bs_prefix->alive_workers;
+                        my $new_state = $cleanup_sub->($channel_name, $state, $alive_workers);
+
+                        return undef, $new_state;
+                    }, no_ban => 1);
                 }
 
-                # 3. delete worker from :heartbeats
-                $c->$bs_prefix->lock_state($WORKERS_CHANNEL, sub ($state) {
-                    delete $state->{$worker->{uuid}};
-                    return undef, $state;
-                }, no_ban => 1);
+                $db->update('guards',
+                    { counter => \"counter - $guards_counter" },
+                    {
+                        worker_uuid => $worker_uuid,
+                        channel_id => $channel->{id},
+                    },
+                );
+                $db->delete('guards',
+                    {
+                        worker_uuid => $worker_uuid,
+                        channel_id  => $channel->{id},
+                        counter     => {'<=', 0},
+                    },
+                );
+                $txn->commit;
             }
+
+            # 3. delete worker from :heartbeats
+            $c->$bs_prefix->lock_state($WORKERS_CHANNEL, sub ($state) {
+                delete $state->{$worker->{uuid}};
+                return undef, $state;
+            }, no_ban => 1);
         }
         my $duration = get_time() - $time;
         $c->app->log->warn('Channels cleanup took ' . (0 + sprintf("%.2f", $duration)) . ' seconds')
@@ -654,7 +688,8 @@ sub register ($self, $app, $config) {
 
     # system channels
     {
-        $app->$bs_prefix->create_channel($WORKERS_CHANNEL, {});
+        $app->$bs_prefix->create_channel($WORKERS_CHANNEL, {}, { keep_events => 0 });
+        $app->$bs_prefix->create_channel($CLEANUP_CHANNEL, { last_cleanup => 0 }, { keep_events => 0 });
         $app->$bs_prefix->global_cleanup();
     }
 }
